open Term
open Overlap
open Rewriting
open Iroiro
open Arg
open Format

type result = YES | NO | MAYBE

(* Global variables *)

let filenames  = ref []
let t          = ref None 
let tt         = ref "ttt2"
let smt        = ref "minismt"


(* Confleunce criteria *)

let kb rs = Wcr.wcr_kb !tt rs

let orthogonal rs =
  Rules.left_linear rs && 
  non_overlapping rs 

let overlay rs = 
  List.for_all (fun (_,p,_,_) -> p = []) (overlap rs)

let weakly_orthogonal rs =
  Rules.left_linear rs && 
  List.for_all (fun (s, t) -> s = t) (cp rs)

let parallel_closed rs (s, t) = List.mem t (rewrite_parallel rs s)

let llpc rs =
  Rules.left_linear rs && 
  List.for_all (parallel_closed rs) (cp rs)

let dc n rs =
  Rules.left_linear rs &&
  Wcr.wcr_n n rs && 
  Sn.sn !tt (cps rs) rs 

let dc2 n rs =
  let p (s, t) = s <> t in
  Rules.left_linear rs &&
  Wcr.wcr_n n rs && 
  Sn.sn !tt (cps ~p rs) rs

let dc3 n rs =
  let p (s, t) = not (parallel_closed rs (s, t)) in
  Rules.left_linear rs &&
  Wcr.wcr_n n rs && 
  Sn.sn !tt (cps ~p rs) rs

let (?:) r v = r := Some v
let (!:) r v = r := v
let prove = ref kb

let options = align [ 
  "-kb", 
     Unit (fun () -> prove := kb), 
     " Knuth-Bendix 1970";
  "-orth",
     Unit (fun () -> prove := orthogonal), 
     " Orthogonal";
  "-overlay",
     Unit (fun () -> prove := overlay), 
     " Overlay";
  "-wo",
     Unit (fun () -> prove := weakly_orthogonal), 
     " Weakly orthogonal";
  "-llpc",
     Unit (fun () -> prove := llpc), 
     " Left-linear parallel closed, Huet 80";
  "-dc",   
     Int (fun n -> prove := dc n), 
     "<n> Decreasingly confluent with n-step joinability ";
  "-dc2",   
     Int (fun n -> prove := dc2 n), 
     "<n> Decreasingly confluent (wrt non-trivial CP) with n-step joinability";
  "-dc3",   
     Int (fun n -> prove := dc3 n), 
     "<n> Decreasingly confluent (wrt non-parallel-closed CP) with n-step joinability";
  "-rl",   
     Int (fun n -> prove := Rl.solve ~tool:!smt n), 
     "<n> Rule labeling with n-step joinability";
  "-rlc",   
     Int (fun n -> prove := Rlc.solve ~tool:!smt n), 
     "<n> Rule labeling with n-step convertibility";
  "-t",    Float  ?:t, "<f> second timeout"; 
  "-tt",   String !:tt, "<tool> set termination tool (default: ttt2)"; 
  "-smt",  String !:smt, "<tool> set smt solver for QF_NIA (default: minismt)"; 

  "-linear", 
     Unit (fun () -> prove := Rules.linear), 
     " check if TRS is linear";
  "-ll", 
     Unit (fun () -> prove := Rules.left_linear), 
     " check if TRS is left-linear";
  "-wcr", 
     Int (fun n -> prove := Wcr.wcr_n n), 
     "<n> check if TRS is WCR by n-step joinability";
] 
     
let short_usage = "Usage: saigawa <options> <file>\n"

let usage = "Usage: saigawa <options> <file>\nOptions are:"


let () = 
  parse options (fun it -> filenames := !filenames @ [it] ) usage;
  if List.length !filenames <> 1 then
    (eprintf "%s%!" short_usage; exit 1)
  else
    let filename = List.hd !filenames in
    let f_timeout () = printf "TIMEOUT@." in
    let f_prove () =
      let rs = Read.read_trs filename in
      if not (Rules.variable_condition rs) then
	(eprintf "Cannot handle TRS with extra variables.@."; exit 1);
      if !prove rs then 
	printf "YES@." 
      else 
	printf "MAYBE@." in
    Alarm.try_with_timeout ?timeout:!t f_timeout f_prove ()
